# frozen_string_literal: true

require_relative 'internal_affairs/method_name_equal'
require_relative 'internal_affairs/node_destructuring'
require_relative 'internal_affairs/node_type_predicate'
require_relative 'internal_affairs/offense_location_keyword'
require_relative 'internal_affairs/redundant_message_argument'
require_relative 'internal_affairs/redundant_location_argument'
require_relative 'internal_affairs/useless_message_assertion'
